import toml
from invoke import task
import glob
import os

pytom = toml.load("pyproject.toml")
package_name = pytom["project"]["name"]
author_name = " - ".join(pytom["project"]["authors"][0]["name"])

doc_dir_name = "docs"
doc_notebooks_dir = "notebooks"


@task
def build(c):

    c.run(f"rm -rf {doc_dir_name}")
    c.run(
        'sphinx-quickstart {} -p {} -a "{}" -q --ext-autodoc'.format(
            doc_dir_name, package_name, author_name
        )
    )

    doc_notebooks = sorted(glob.glob("{}/*.ipynb".format(doc_notebooks_dir)))
    c.run(
        "jupyter nbconvert --to rst {} --output-dir={}/notebooks".format(
            " ".join(doc_notebooks), doc_dir_name
        )
    )
    toc_nbs = [
        "   notebooks/{}".format(np.split("/")[-1].split(".")[0])
        for np in doc_notebooks
    ]

    index_rst = """
Welcome to {}'s documentation!
=====================================================================================

.. toctree::
   :maxdepth: 2
   :caption: Contents:

   autosumm
{}
   {}

Indices and tables
==================

* :ref:`genindex`
* :ref:`modindex`
* :ref:`search`
""".format(
        package_name, "\n".join(toc_nbs), "release_notes/main"
    )

    autosumm_rst = """

API
===

.. automodapi:: {}

""".format(
        package_name
    )

    with open(os.path.join(doc_dir_name, "index.rst"), "w") as fp:
        fp.write(index_rst)

    with open(os.path.join(doc_dir_name, "autosumm.rst"), "w") as fp:
        fp.write(autosumm_rst)

    c.run("cp -r docs_config/* docs/")
    c.run("pip install -r docs/requirements.txt")
    c.run("sphinx-build docs docs/_build")



import json
import os
import random
from typing import Optional

from jkg_evaluators.util import zip_dir


def dump_separated_minor_nb(
    notebook_path: str,
    output_directory: str,
    members: list,
    task_name: str,
    zip_dir_path: Optional[str] = None,
):
    nb_dic = json.load(open(notebook_path))
    cells = nb_dic.pop("cells")
    task_cell_lists = []
    for idx, cell in enumerate(cells):
        if cell["source"][0].startswith("####"):
            start_ind = idx
            end_ind = idx + 4
            task_cell_lists.append(cells[start_ind:end_ind])
    random.shuffle(members)
    os.makedirs(output_directory, exist_ok=True)

    for idx, member in enumerate(members):
        member_path = os.path.join(output_directory, member)
        os.makedirs(member_path, exist_ok=True)
        member_nb_path = os.path.join(
            member_path, "{}.ipynb".format(task_name)
        )
        member_nb = nb_dic.copy()
        member_nb["cells"] = task_cell_lists[idx]
        json.dump(member_nb, open(member_nb_path, "w"))

    if zip_dir_path is not None:
        zip_dir(zip_dir_path, output_directory, task_name)



import os
import zipfile
from typing import Optional


def zipdir_walk(path, ziph):
    for root, dirs, files in os.walk(path):
        for file in files:
            ziph.write(os.path.join(root, file))


def get_code_nb_cell(
    code="", metadata=None, outputs=None, execution_count=None
):
    return {
        "cell_type": "code",
        "source": [code],
        "metadata": metadata or {},
        "outputs": outputs or [],
        "execution_count": execution_count,
    }


def get_md_nb_cell(md="", metadata=None):
    return {
        "cell_type": "markdown",
        "source": [md],
        "metadata": metadata or {},
    }


def zip_dir(
    zip_dir_path: str, dir_to_zip: str, zip_file_name: Optional[str] = None
):
    zip_file_path = os.path.join(zip_dir_path, "{}.zip".format(zip_file_name))
    zipf = zipfile.ZipFile(zip_file_path, "w", zipfile.ZIP_DEFLATED)
    zipdir_walk(dir_to_zip, zipf)
    zipf.close()



from typing import Callable

from jkg_evaluators.core import CasePerformance, CompleteEvaluation, EvalCase


class DragonFindCase(EvalCase):
    performance_bigger_better = False
    main_complexity_var = "number_of_cows"

    def __init__(self, number_of_cows: int, dragon_coming_to: int):
        self.number_of_cows = number_of_cows
        self.dragon_coming_to = dragon_coming_to
        self.looks = 0

    def _evaluate(self, solution: Callable) -> CasePerformance:

        out = solution(self.is_dead, self.number_of_cows)

        is_success = out == self.dragon_coming_to

        return CasePerformance(
            is_successful=is_success, performance=self.looks
        )

    def is_dead(self, k):
        self.looks += 1
        return k < self.dragon_coming_to


def get_dragon_kwargs():
    dragon_kwargs_list = []

    for cownum in range(10, 501):
        for _dragon_coming_to in range(1, cownum):
            dragon_kwargs_list.append(
                {
                    "number_of_cows": cownum,
                    "dragon_coming_to": _dragon_coming_to,
                }
            )
    return dragon_kwargs_list


dragonfind_10_to_500 = CompleteEvaluation(
    get_case_kwarg_list=get_dragon_kwargs, case=DragonFindCase
)



from invoke import task
from .vars import package_name, boilerplate_branch


@task
def lint(c):
    c.run("black . -l 79 --exclude \.*venv")
    c.run(f"isort -rc {package_name} -m 3 -tc")
    c.run(f"flake8 {package_name}")


@task
def update_boilerplate(c):  # TODO: maybe drop template package folder
    c.run("git fetch boilerplate")  # TODO: this knows the name of the remote
    c.run(f"git merge boilerplate/{boilerplate_branch} --no-edit")


@task
def notebook(c):
    c.run("python invoke_commands/notebook_runner.py")



import random
import string
from typing import List


def get_word_lists() -> List[dict]:
    word_lists = [
        {"list_of_words": ["b", "bb", "bBb"]},
        {"list_of_words": ["aa", "aaa", "aa", "a"]},
        {"list_of_words": ["ba", "babab", "AaAaA", "Ahha"]},
        {"list_of_words": ["123", "lala", "", "", "", "aA"]},
    ]
    random.seed(42069)
    for _r in range(10, 501):
        word_lists.append(
            {
                "list_of_words": [
                    "".join(
                        random.choices(
                            string.ascii_letters + string.digits,
                            k=random.randint(0, 120),
                        )
                    )
                    for _ in range(_r)
                ]
            }
        )
    return word_lists


def get_num_lists() -> List[dict]:

    num_lists = [
        {"list_of_numbers": [0, 1, 2, 3, 4]},
        {"list_of_numbers": [0, 0, 10, 3, 10]},
        {"list_of_numbers": [-20, -40, 0, 1, 2, 3, 40]},
        {"list_of_numbers": [-1100, 100, 20]},
        {"list_of_numbers": [0, 1]},
        {"list_of_numbers": [21, 120, 220]},
    ]

    random.seed(42069)
    for _r in range(10, 501):

        num_lists.append(
            {
                "list_of_numbers": [
                    random.randint(-1000, 1000) for _ in range(_r)
                ]
            }
        )

    return num_lists


def get_words_with_letter() -> List[dict]:
    word_lists = get_word_lists()
    return [
        {"letter": random.choice(string.ascii_letters), **kwargs}
        for kwargs in word_lists
    ]


def get_numberlists_with_numbers() -> List[dict]:
    num_lists = get_num_lists()
    return [
        {"number": random.randint(1, 10), **kwargs} for kwargs in num_lists
    ]



import json
import os
import random
import shutil

import numpy as np
import pandas as pd

from .comparator_class import SolutionComparator


def _create_hotel_input(size):
    return [
        {
            "lat": (random.betavariate(3, 3) - 0.5) * 125,
            "lon": (random.betavariate(3, 3) - 0.5) * 300,
        }
        for _ in range(size)
    ]


def _create_hotel_filter_input(size):
    return [
        {
            "lat": (random.betavariate(3, 3) - 0.5) * 125,
            "lon": (random.betavariate(3, 3) - 0.5) * 300,
            "stars": np.linspace(0, 5, 11)[np.random.randint(11)],
            **_get_price_pair(),
        }
        for _ in range(size)
    ]


def _get_price_pair():
    prices = np.sort(np.random.exponential(70, size=2))
    return {"min_price": prices[0], "max_price": prices[1]}


class HotelSolutionComparator(SolutionComparator):
    @staticmethod
    def _get_input(size):
        return _create_hotel_input(size)

    @staticmethod
    def _dump_data(size, path):
        data_path = os.path.join("data", f"{size}.csv")
        shutil.copy(data_path, path)


class HotelFilterSolutionComparator(HotelSolutionComparator):
    @staticmethod
    def _get_input(size):
        return _create_hotel_filter_input(size)


def get_hotel_data(data_root="data"):
    os.makedirs(data_root, exist_ok=True)
    for data_size_k in [10, 20, 50, 100, 200, 500]:
        data_size = data_size_k * 1000
        dl_path = (
            "https://borza-hotelcom-data.s3.eu-central-1"
            f".amazonaws.com/challenge-{data_size}.csv"
        )
        df = pd.read_csv(dl_path)
        write_path = os.path.join(data_root, f"{data_size}.csv")
        df.to_csv(write_path, index=False)


def dump_hotel_input(size, path="inputs.json"):
    obj = _create_hotel_input(size)
    with open(path, "w") as fp:
        json.dump(obj, fp)


def dump_hotel_filter_input(size, path="inputs.json"):
    obj = _create_hotel_filter_input(size)
    with open(path, "w") as fp:
        json.dump(obj, fp)



from invoke import task

from .vars import package_name


@task
def clean(
    c,
    docs=False,
    build=False,
    bytecode=False,
    test=False,
    sonar=False,
    all=False,
):
    patterns = []
    if docs or all:
        patterns.append("docs")
    if build or all:
        patterns += ["build", "{}.egg-info".format(package_name)]
    if bytecode or all:
        patterns.append("**/*.pyc")
    if sonar or all:
        patterns += [
            "{}/.sonar".format(package_name),
            "{}/.scannerwork".format(package_name),
            "{}/sonar-project.properties".format(package_name),
        ]
    if test or all:
        patterns += [
            ".pytest_cache",
            ".coverage",
            "**/.coverage",
            "htmlcov",
            "**/coverage.xml",
        ]

    for pattern in patterns:
        c.run("rm -rf {}".format(pattern))


@task
def prune(c):
    clean(c, all=True)



from typing import Callable

from jkg_evaluators.core import CasePerformance, CompleteEvaluation, EvalCase


class BasicEggCase(EvalCase):

    performance_bigger_better = False
    main_complexity_var = "max_floor"

    def __init__(self, max_floor: int, egg_count: int, floor_count: int):
        self.max_floor = max_floor
        self.egg_count = egg_count
        self.floor_count = floor_count
        self.attempts = 0
        self.broken_eggs = 0
        self.attempted_after_ran_out = False

    def _evaluate(self, solution: Callable) -> CasePerformance:

        out = solution(self.throw_attempt)

        is_success = (out == self.max_floor) and (
            not self.attempted_after_ran_out
        )

        return CasePerformance(
            is_successful=is_success, performance=self.attempts
        )

    def throw_attempt(self, k):
        if self.broken_eggs == self.egg_count:
            self.attempted_after_ran_out = True
        self.attempts += 1
        if k <= self.max_floor:
            return False
        else:
            self.broken_eggs += 1
            return True


def get_egg_kwargs():
    return [
        {"egg_count": 2, "floor_count": 100, "max_floor": i}
        for i in range(0, 101)
    ]


eggdrop_100floor_2egg = CompleteEvaluation(
    get_case_kwarg_list=get_egg_kwargs,
    case=BasicEggCase,
)



import os
from invoke import task
import io

from .vars import mymodule


@task
def new(c):

    version = mymodule.__version__

    c.run("python setup.py sdist")
    c.run("twine check dist/*")
    twine_pw = os.environ.get("TWINE_PASSWORD")
    if twine_pw:
        c.run(
            f"twine upload dist/*{version}.tar.gz -u __token__ -p $TWINE_PASSWORD"
        )
    else:
        c.run(f"twine upload dist/*{version}.tar.gz")


@task
def tag(c):

    version = mymodule.__version__

    f = io.StringIO()
    c.run("git rev-parse --abbrev-ref HEAD", out_stream=f)
    branch = f.getvalue().strip()
    f.close()

    if branch == "master":
        tag_version = "v{}".format(version)
        f2 = io.StringIO()
        c.run("git tag", out_stream=f2)
        tags = f2.getvalue().split()
        print(tags)
        if tag_version not in tags:
            current_release_path = "docs_config/current_release.rst"
            with open(current_release_path) as fp:
                notes = fp.read()
            with open(
                "docs_config/release_notes/{}.rst".format(tag_version), "w"
            ) as fp:
                fp.write(notes)
            c.run(f"git tag -a {tag_version} -m '{notes}'")
            with open(current_release_path, "w") as fp:
                fp.write("")
            c.run("git push --tags")
        else:
            print("{} version already tagged".format(tag_version))
    else:
        print("only master branch can be tagged")



import random
from copy import deepcopy
from typing import Callable, List, Optional, Type, Union


class EvaluationCaseResult:
    def __init__(
        self,
        is_successful: bool = False,
        performance: Optional[Union[float, int]] = None,
        e: Optional[Exception] = None,
    ):

        self.is_successful = is_successful
        self.performance = performance
        self.exception_thrown = e


class CasePerformance:
    def __init__(
        self,
        is_successful: bool = False,
        performance: Optional[Union[float, int]] = None,
    ):

        self.is_successful = is_successful
        self.performance = performance


class EvalCase:

    performance_bigger_better = False
    main_complexity_var = None

    def _evaluate(self, solution: Callable) -> CasePerformance:

        return CasePerformance()

    def evaluate(self, solution: Callable) -> EvaluationCaseResult:

        try:
            out = self._evaluate(solution)
            e = None
        except Exception as _e:
            e = _e
            out = CasePerformance()

        return EvaluationCaseResult(
            is_successful=out.is_successful, e=e, performance=out.performance
        )

    def describe_case(self):
        pass


class CompleteEvaluation:
    def __init__(
        self,
        get_case_kwarg_list: Callable[[], List[dict]],
        case: Type[EvalCase],
    ):

        self.case = case
        self.case_kwarg_list_getter = get_case_kwarg_list
        self.case_kwarg_list = []

    def _load(self):
        self.case_kwarg_list = self.case_kwarg_list_getter()
        self.case_results_list = [EvaluationCaseResult()] * len(
            self.case_kwarg_list
        )
        self.success_list = [False] * len(self.case_kwarg_list)
        self.error_list = [None] * len(self.case_kwarg_list)
        self.performance_list = [None] * len(self.case_kwarg_list)

    def _run_all(self, solution: Callable):

        num_kwargs = len(self.case_kwarg_list)
        self.case_results_list = [EvaluationCaseResult()] * num_kwargs
        self.success_list = [False] * num_kwargs
        self.error_list = [None] * num_kwargs
        self.performance_list = [None] * num_kwargs

        iterable = random.sample(
            list(enumerate(self.case_kwarg_list)), num_kwargs
        )

        for kwarg_idx, case_kwargs in iterable:
            copied_kwargs = deepcopy(case_kwargs)
            case = self.case(**copied_kwargs)
            act_case_results = case.evaluate(solution)
            self.case_results_list[kwarg_idx] = act_case_results

            if act_case_results.exception_thrown is not None:
                self.error_list[kwarg_idx] = act_case_results.exception_thrown
            self.success_list[kwarg_idx] = act_case_results.is_successful
            self.performance_list[kwarg_idx] = act_case_results.performance

    def _get_str_results(self) -> str:

        numeric_performances = [
            p for p in self.performance_list if p is not None
        ]

        if len(numeric_performances) > 0:

            if self.case.performance_bigger_better:
                bestfun = max
                worstfun = min
            else:
                bestfun = min
                worstfun = max

            performance_summas = [
                "- best performance: {}".format(bestfun(numeric_performances)),
                "- worst performance: {}".format(
                    worstfun(numeric_performances)
                ),
                "- mean performance: {}".format(
                    sum(numeric_performances) / len(numeric_performances)
                ),
            ]
        else:
            performance_summas = []

        error_strings = []
        miss_strings = []
        for idx, succ in enumerate(self.success_list):

            if not succ:
                e = self.error_list[idx]
                act_kwargs = self.case_kwarg_list[idx]
                if e is not None:
                    error_strings.append(
                        "\n\n ERROR at: \n {} \n - {} ({})".format(
                            act_kwargs, type(e).__name__, e
                        )
                    )
                else:
                    miss_strings.append(
                        "\n\n BAD SOLUTION at: \n {}".format(act_kwargs)
                    )

        return "\n".join(
            [
                "- success rate: {}/{} ({}%)".format(
                    sum(self.success_list),
                    len(self.case_kwarg_list),
                    round(
                        sum(self.success_list) / len(self.case_kwarg_list), 2
                    )
                    * 100,
                ),
                "- error count: {}".format(
                    sum([e is not None for e in self.error_list])
                ),
                *performance_summas,
                *miss_strings[:10],
                *error_strings[:10],
            ]
        )

    def _get_performance_plot(self) -> dict:

        if self.case.main_complexity_var is None:
            return {}
        else:
            cvar = self.case.main_complexity_var
            successes = []
            fails = []
            errors = []
            c_values = []
            for idx, case_kwargs in enumerate(self.case_kwarg_list):

                perf_value = self.performance_list[idx]
                if self.success_list[idx]:
                    successes.append(perf_value)
                    fails.append(None)
                    errors.append(None)
                elif self.error_list[idx] is not None:
                    successes.append(None)
                    fails.append(None)
                    errors.append(0)
                else:
                    successes.append(None)
                    fails.append(0)
                    errors.append(None)

                try:
                    c_values.append(case_kwargs[cvar])
                except KeyError:
                    c_values.append(idx)

        out = {cvar: c_values}
        for ser, label in [
            (successes, "Successes"),
            (fails, "Bad solutions"),
            (errors, "Errors"),
        ]:
            if sum([e is None for e in ser]) < len(ser):
                out[label] = ser
        return out

    def evaluate(self, solution: Callable, return_output=False):
        if not self.case_kwarg_list:
            self._load()
        self._run_all(solution)
        print(self)

        if return_output:
            success_rate = sum(self.success_list) / len(self.case_kwarg_list)
            return success_rate == 1

    def visualize(self, solution: Callable):

        import matplotlib.pyplot as plt

        self._run_all(solution)

        data = self._get_performance_plot()
        plt.figure(figsize=(11, 7))

        for label, ser in data.items():
            if label != self.case.main_complexity_var:
                if label == "-Successes":
                    marker = None
                else:
                    marker = "o"
                plt.plot(
                    self.case.main_complexity_var,
                    label,
                    marker=marker,
                    linestyle="",
                    data=data,
                )
        plt.ylabel("performance")
        plt.xlabel(self.case.main_complexity_var)
        plt.legend()
        plt.show()

    def __str__(self):

        if len(self.case_results_list) > 0:
            return self._get_str_results()
        else:
            return "not yet evaluated"

    def __repr__(self):

        return self.__str__()



from jupyter_client.kernelspec import KernelSpecManager
from notebook.notebookapp import main


class MyManager(KernelSpecManager):
    def get_kernel_spec(self, kernel_name):
        init_resp = super().get_kernel_spec(kernel_name)
        init_resp.argv = [
            *init_resp.argv,
            """--IPKernelApp.exec_lines=['import sys', 'sys.path.append("..")']""",
        ]
        return init_resp


if __name__ == "__main__":
    main(kernel_spec_manager_class=MyManager)



import os
from invoke import task

from .vars import package_name


@task
def setup(c):
    c.run("docker run -d --name sonarqube -p 9000:9000 -p 9092:9092 sonarqube")


@task
def scan(c):

    with open(
        os.path.join(package_name, "sonar-project.properties"), "w"
    ) as fp:
        fp.write(
            "sonar.projectKey={}\n"
            "sonar.python.coverage.reportPaths=coverage.xml\n"
            "sonar.scm.disabled=true".format(package_name)
        )

    c.run(
        "docker run -e SONAR_HOST_URL={} "
        '--user="$(id -u):$(id -g)" '
        '-t -v "{}/{}:/usr/src" sonarsource/sonar-scanner-cli'.format(
            "http://172.17.0.2:9000", os.getcwd(), package_name
        )
    )


@task
def kill(c):
    c.run("docker kill sonarqube")
    c.run("docker container rm sonarqube")



import itertools
import json
import os
import shutil
import subprocess
import time

import numpy as np
import pandas as pd
import yaml


class SolutionComparator:
    def __init__(self, data_sizes, input_sizes, solutions_to_compare="all"):

        if solutions_to_compare == "all":
            self.solutions_to_compare = next(
                os.walk(os.path.join("..", "solutions"))
            )[1]
        elif isinstance(solutions_to_compare, str):
            self.solutions_to_compare = [solutions_to_compare]
        else:
            self.solutions_to_compare = solutions_to_compare

        self.input_dicts = {
            d_size: {isize: self._get_input(isize) for isize in input_sizes}
            for d_size in data_sizes
        }
        self.comparison_data = []
        self._tmp_dir = "tmp"

    def _cleanup(self):
        shutil.rmtree(self._tmp_dir, ignore_errors=True)

    @staticmethod
    def _get_input(size):
        return [size]

    @staticmethod
    def _dump_data(size, data_path):
        with open(data_path, "w") as fp:
            fp.write(f"{size} nothing")

    def _try_running_command(
        self, command_dic: dict, command_name: str, act_solution: str
    ):
        command = command_dic.get(command_name)
        if command:
            print(f"running {command_name} for {act_solution}")

            command_start_time = time.time()
            _proc = subprocess.Popen(
                command.split(),
                # stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=self._tmp_dir,
            )
            _proc.wait()
            _stdout, _stderr = _proc.communicate()
            command_time = time.time() - command_start_time
            if _proc.returncode != 0:
                self._cleanup()
                print(f"\n\n{act_solution} {command_name} ERROR:\n--")
                print(_stderr.decode("utf-8"))
                raise RuntimeError(_stderr.decode("utf-8"))

        else:
            print(f"no {command_name} for {act_solution}")
            command_time = None

        return command_time

    def run_comparison(self):

        for act_solution in self.solutions_to_compare:
            print(f"solution: {act_solution}")
            self._cleanup()
            act_solution_path = os.path.join("..", "solutions", act_solution)
            shutil.copytree(act_solution_path, self._tmp_dir)
            try:
                commands = yaml.safe_load(
                    open(os.path.join("tmp", "commands.yaml"))
                )
            except FileNotFoundError:
                self._cleanup()
                raise FileNotFoundError(
                    f"no commands.yaml in solution directory: {act_solution}"
                )

            self._try_running_command(
                commands, "setup-env-command", act_solution
            )

            for data_size, input_dict in self.input_dicts.items():

                data_path = os.path.join("data", f"{data_size}.csv")
                shutil.copy(data_path, os.path.join(self._tmp_dir, "data.csv"))

                self._try_running_command(
                    commands, "etl-command", act_solution
                )

                for input_size, input_object in input_dict.items():
                    print(f"input size: {input_size} - data size: {data_size}")

                    json.dump(
                        input_object,
                        open(os.path.join(self._tmp_dir, "inputs.json"), "w"),
                    )

                    calc_time = self._try_running_command(
                        commands, "process-command", act_solution
                    )
                    if calc_time is None:
                        self._cleanup()
                        raise ValueError(
                            "no process-command specified "
                            f"in {act_solution} commands.yaml"
                        )

                    try:
                        output_list = json.load(
                            open(os.path.join(self._tmp_dir, "outputs.json"))
                        )
                    except FileNotFoundError:
                        self._cleanup()
                        raise RuntimeError(
                            f"solution {act_solution} "
                            "did not produce an outputs.json"
                        )

                    self.comparison_data.append(
                        {
                            "calc_time": calc_time,
                            "input_size": input_size,
                            "data_size": data_size,
                            "solution": act_solution,
                            "output_list": output_list.copy(),
                        }
                    )

                self._try_running_command(
                    commands, "cleanup-command", act_solution
                )

        self._cleanup()

    def plot_comparison(self):

        if self.comparison_data:
            pd.DataFrame(self.comparison_data).pivot_table(
                index=["data_size", "input_size"],
                columns="solution",
                values="calc_time",
            ).plot.bar(figsize=(13, 7))
        else:
            print("run the comparison first")

    def output_comparison(self):

        if self.comparison_data:
            return (
                pd.DataFrame(self.comparison_data)
                .groupby("solution")
                .apply(
                    lambda df: pd.DataFrame(
                        itertools.chain(*df["output_list"].tolist())
                    ).reset_index(drop=True)
                )
                .fillna(0)
                .pipe(
                    lambda df: df.reset_index().pivot_table(
                        columns="solution",
                        index=[
                            "level_1",
                            *df.select_dtypes(exclude=np.number).columns,
                        ],
                    )
                )
            )
        else:
            print("run the comparison first")



import os
import glob
import json
from invoke import task

from .vars import package_name, doc_notebooks_dir


@task
def test(c, option="", html=False, xml=False, notebook_tests=False):

    comm = "python -m pytest --cov={}".format(package_name)
    if option:
        comm += " --{}".format(option)
    if html:
        comm += " --cov-report=html"
    elif xml:
        comm += f" --cov-report=xml:{package_name}/coverage.xml"

    if notebook_tests:
        new_test_scripts = []
        for nb_idx, nb_file in enumerate(
            glob.glob(os.path.join(doc_notebooks_dir, "*.ipynb"))
        ):
            nb_dic = json.load(open(nb_file))
            nb_code = "\n".join(
                [
                    "\n".join(c["source"])
                    for c in nb_dic["cells"]
                    if (c["cell_type"] == "code")
                ]
            )
            if len(nb_code) > 0:
                new_test_scripts.append(
                    f"def test_nb_integration_{nb_idx}():\n"
                    + "\n".join([f"    {s}" for s in nb_code.split("\n")])
                )

        test_root = os.path.join(package_name, "tests")
        if not os.path.exists(test_root):
            os.makedirs(test_root)

        test_notebook_path = os.path.join(test_root, "test_nb_integrations.py")
        with open(test_notebook_path, "w") as fp:
            fp.write("\n\n".join(new_test_scripts))

    c.run(comm)
    c.run("rm -f {}/tests/test_nb_integrations.py".format(package_name))



import toml
import importlib

pytom = toml.load("pyproject.toml")
package_name = pytom["project"]["name"]
author_name = " - ".join(pytom["project"]["authors"][0]["name"])

doc_dir_name = "docs"
doc_notebooks_dir = "notebooks"

mymodule = importlib.import_module(package_name)

boilerplate_branch = "master"



from collections import Counter
from typing import Callable

from jkg_evaluators.core import CasePerformance, CompleteEvaluation, EvalCase
from jkg_evaluators.input_kwarg_generators import (
    get_num_lists,
    get_numberlists_with_numbers,
    get_word_lists,
    get_words_with_letter,
)


class IndexWithMostALetters(EvalCase):
    performance_bigger_better = True

    def __init__(self, list_of_words: list):

        self.list_of_words = list_of_words
        self.n = len(list_of_words)

    def _evaluate(self, solution: Callable) -> CasePerformance:

        true_solution = -1
        top_num_as = 0
        for idx, word in enumerate(self.list_of_words):
            act = sum([_l.lower() == "a" for _l in word])
            if act > top_num_as:
                true_solution = idx
                top_num_as = act

        out = solution(self.list_of_words)

        is_success = out == true_solution

        return CasePerformance(
            is_successful=is_success, performance=int(is_success)
        )


class LetterOccurrences(EvalCase):
    performance_bigger_better = True

    def __init__(self, list_of_words: list, letter: str):

        self.list_of_words = list_of_words
        self.letter = letter

    def _evaluate(self, solution: Callable) -> CasePerformance:

        true_solution = 0
        for word in self.list_of_words:
            if self.letter.lower() in word.lower():
                true_solution += 1

        out = solution(self.list_of_words, self.letter)

        is_success = out == true_solution

        return CasePerformance(
            is_successful=is_success, performance=int(is_success)
        )


class WordWithMostOfLetter(EvalCase):
    performance_bigger_better = True

    def __init__(self, list_of_words: list, letter: str):

        self.list_of_words = list_of_words
        self.letter = letter

    def _evaluate(self, solution: Callable) -> CasePerformance:

        true_solution = ""
        max_num = 0
        for word in self.list_of_words:
            letter_count = sum(
                [ll == self.letter.lower() for ll in word.lower()]
            )
            if letter_count > max_num:
                max_num = letter_count
                true_solution = word

        out = solution(self.list_of_words, self.letter)

        is_success = out == true_solution

        return CasePerformance(
            is_successful=is_success, performance=int(is_success)
        )


class LargestMultiple(EvalCase):
    performance_bigger_better = True

    def __init__(self, list_of_numbers: list):

        self.list_of_numbers = list_of_numbers

    def _evaluate(self, solution: Callable) -> CasePerformance:

        sorted_nums = sorted(self.list_of_numbers)

        out = solution(self.list_of_numbers)

        bot_multi = sorted_nums[-1] * sorted_nums[-2]
        top_multi = sorted_nums[0] * sorted_nums[1]

        if bot_multi > top_multi:
            true_solution = bot_multi
        else:
            true_solution = top_multi

        is_success = out == true_solution

        return CasePerformance(
            is_successful=is_success, performance=int(is_success)
        )


class SumOfDistinctOddPosInts(EvalCase):
    performance_bigger_better = True

    def __init__(self, list_of_numbers: list):

        self.list_of_numbers = list_of_numbers

    def _evaluate(self, solution: Callable) -> CasePerformance:

        true_solution = 0
        for x in set(self.list_of_numbers):
            if ((x % 2) != 0) and (x > 0):
                true_solution += x

        out = solution(self.list_of_numbers)

        is_success = out == true_solution

        return CasePerformance(
            is_successful=is_success, performance=int(is_success)
        )


class LargestAscending(EvalCase):
    performance_bigger_better = True

    def __init__(self, list_of_numbers: list):

        self.list_of_numbers = list_of_numbers

    def _evaluate(self, solution: Callable) -> CasePerformance:
        def check_sorted(num):
            sl = sorted(str(abs(num)))
            return "".join(sl) == str(abs(num))

        filtered = [x for x in self.list_of_numbers if check_sorted(x)]

        try:
            true_solution = max(filtered)
        except ValueError:
            true_solution = 0

        out = solution(self.list_of_numbers)

        is_success = out == true_solution

        return CasePerformance(
            is_successful=is_success, performance=int(is_success)
        )


class SmallestWhereDoubleAlso(EvalCase):
    performance_bigger_better = True

    def __init__(self, list_of_numbers: list):

        self.list_of_numbers = list_of_numbers

    def _evaluate(self, solution: Callable) -> CasePerformance:

        filtered = [
            x
            for x in self.list_of_numbers
            if ((x * 2) in self.list_of_numbers)
        ]

        try:
            true_solution = min(filtered)
        except ValueError:
            true_solution = 0

        out = solution(self.list_of_numbers)

        is_success = out == true_solution

        return CasePerformance(
            is_successful=is_success, performance=int(is_success)
        )


class LargestEvenDivided(EvalCase):
    performance_bigger_better = True

    def __init__(self, list_of_numbers: list, number: int):

        self.list_of_numbers = list_of_numbers
        self.number = number

    def _evaluate(self, solution: Callable) -> CasePerformance:

        true_solution = None

        for n in self.list_of_numbers:
            if (n / self.number) % 2 == 0:
                if true_solution is None:
                    true_solution = n
                elif n > true_solution:
                    true_solution = n

        if true_solution is None:
            true_solution = 0

        out = solution(self.list_of_numbers, self.number)

        is_success = out == true_solution

        return CasePerformance(
            is_successful=is_success, performance=int(is_success)
        )


class LastWithThreeMultDiff(EvalCase):
    performance_bigger_better = True

    def __init__(self, list_of_numbers: list, number: int):

        self.list_of_numbers = list_of_numbers
        self.number = number

    def _evaluate(self, solution: Callable) -> CasePerformance:

        true_solution = 0

        for n in self.list_of_numbers:
            if (n - self.number) % 3 == 0:
                true_solution = n

        out = solution(self.list_of_numbers, self.number)

        is_success = out == true_solution

        return CasePerformance(
            is_successful=is_success, performance=int(is_success)
        )


class LongestWithThreeSame(EvalCase):
    performance_bigger_better = True

    def __init__(self, list_of_words: list):
        self.list_of_words = list_of_words

    def _evaluate(self, solution: Callable) -> CasePerformance:

        true_solution = ""

        for w in self.list_of_words:
            if (max(Counter(w.lower()).values(), default=0) >= 3) and (
                len(w) > len(true_solution)
            ):
                true_solution = w

        out = solution(self.list_of_words)
        is_success = out == true_solution

        return CasePerformance(
            is_successful=is_success, performance=int(is_success)
        )


string_with_most_a_letters = CompleteEvaluation(
    get_case_kwarg_list=get_word_lists, case=IndexWithMostALetters
)

letter_occurrences = CompleteEvaluation(
    get_case_kwarg_list=get_words_with_letter, case=LetterOccurrences
)

word_with_most_of_letters = CompleteEvaluation(
    get_case_kwarg_list=get_words_with_letter, case=WordWithMostOfLetter
)

largest_multiple = CompleteEvaluation(
    get_case_kwarg_list=get_num_lists, case=LargestMultiple
)

sum_odd_positives = CompleteEvaluation(
    get_case_kwarg_list=get_num_lists, case=SumOfDistinctOddPosInts
)

largest_ascending_num = CompleteEvaluation(
    get_case_kwarg_list=get_num_lists, case=LargestAscending
)

smallest_where_double_also = CompleteEvaluation(
    get_case_kwarg_list=get_num_lists, case=SmallestWhereDoubleAlso
)

largest_even_divided = CompleteEvaluation(
    get_case_kwarg_list=get_numberlists_with_numbers, case=LargestEvenDivided
)

last_with_three_multiple_difference = CompleteEvaluation(
    get_case_kwarg_list=get_numberlists_with_numbers,
    case=LastWithThreeMultDiff,
)

longest_with_three_same_letters = CompleteEvaluation(
    get_case_kwarg_list=get_word_lists, case=LongestWithThreeSame
)



from ._version import VERSION
from .challenges.dragonfind import dragonfind_10_to_500  # noqa
from .challenges.egg import eggdrop_100floor_2egg  # noqa
from .challenges.minors import (  # noqa
    largest_ascending_num,
    largest_even_divided,
    largest_multiple,
    last_with_three_multiple_difference,
    letter_occurrences,
    longest_with_three_same_letters,
    smallest_where_double_also,
    string_with_most_a_letters,
    sum_odd_positives,
    word_with_most_of_letters,
)

__version__ = VERSION



import itertools
import json
import os
from typing import Optional

import unidecode

from jkg_evaluators.util import get_code_nb_cell, get_md_nb_cell, zip_dir


def distribute_originals(
    original_directory: str,
    members: list,
    task_name: str,
    collective_task_name: str,
    output_directory: Optional[str] = None,
    zip_dir_path: Optional[str] = None,
):

    if output_directory is None:
        output_directory = original_directory
    os.makedirs(output_directory, exist_ok=True)

    mem_task_dic = {}
    for m in members:
        nb_path = os.path.join(
            original_directory, m, "{}.ipynb".format(task_name)
        )
        with open(nb_path) as fp:
            nb_js = json.load(fp)

        for line in itertools.chain(*[c["source"] for c in nb_js["cells"]]):
            if line.startswith("from jkg_evaluators import"):
                mem_task_dic[m] = line.split()[-1]
                break

    import_file_source = "\n".join(
        [
            "from jkg_evaluators import {} as {}".format(
                task, unidecode.unidecode(m)
            )
            for m, task in mem_task_dic.items()
        ]
    )
    import_file_path = os.path.join(output_directory, "__import_file__.py")

    with open(import_file_path, "w") as fp:
        fp.write(import_file_source)

    member_imports = [
        "from __import_file__ import {}".format(unidecode.unidecode(m))
        for m in members
    ]

    import_cell_source = "\n".join(
        ["import sys", "sys.path.insert(0, '..')", *member_imports]
    )

    all_cells = [get_code_nb_cell(import_cell_source)]

    for m in members:
        all_cells += get_member_task_solution_cells(m)

    nb_js["cells"] = all_cells

    for m in members:
        os.makedirs(os.path.join(output_directory, m), exist_ok=True)
        out_nb_path = os.path.join(
            output_directory, m, "{}.ipynb".format(collective_task_name)
        )
        with open(out_nb_path, "w") as fp:
            json.dump(nb_js, fp)

    if zip_dir_path is not None:
        zip_dir(zip_dir_path, output_directory, collective_task_name)


def get_member_task_solution_cells(m):
    dem = unidecode.unidecode(m)
    return [
        get_md_nb_cell("## {}".format(m)),
        get_code_nb_cell(
            "\n".join(
                [
                    "def {}_solution():".format(dem),
                    "    # modify this function",
                    "    # return value and parameters and all",
                    "    return 0",
                ]
            )
        ),
        get_code_nb_cell("{}.evaluate({}_solution)".format(dem, dem)),
    ]
